/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdispatch.h"
#include <gdam/gdam-protocol.h>
#include "gdamdevice.h"
#include "gdamnamespace.h"
#include "gdamevent.h"
#include "gdamexternal.h"
#include "gdamcachebuffer.h"
#include "gdamserverdebug.h"
#include "gdamwavetable.h"
#include <stdlib.h>
#include <stdio.h>

static FILE*    gdam_dispatch_log_fp = NULL;
static gboolean gdam_dispatch_logging = FALSE;
static int      dispatch_logging_indent = 0;

static void print_spaces(FILE* fp, int n)
{
	char buf[129] = {0};
	if (buf[0] == 0) {
		memset(buf, ' ', 128);
	}
	while (n > 0) {
		int spit = n;
		if (spit > 128) spit = 128;
		fputs(buf + 128 - spit, fp);
		n -= spit;
	}
}

void  gdam_dispatch_log(const char* fmt, ...)
{
	va_list args;
	if (!gdam_dispatch_logging || !gdam_dispatch_log_fp)
		return;
	print_spaces(gdam_dispatch_log_fp, dispatch_logging_indent * 2);
	va_start(args, fmt);
	vfprintf(gdam_dispatch_log_fp, fmt, args);
	va_end(args);
	fprintf(gdam_dispatch_log_fp, "\n");
}

void gdam_dispatch_set_log(const char* filename)
{
	if (gdam_dispatch_log_fp) {
		fclose(gdam_dispatch_log_fp);
		gdam_dispatch_log_fp = NULL;
	}
	if (filename) {
		gdam_dispatch_log_fp = fopen(filename, "a");
		if (!gdam_dispatch_log_fp) {
			g_warning("cannot open dispatch logfile %s", filename);
		} else {
			setbuf(gdam_dispatch_log_fp, 0);
			gdam_dispatch_logging = TRUE;
		}
	}
}


/*
 * Builtin command dispatchers.
 *
 * Each of these is a CommandDispatcher for a particular
 * message type.
 */

static GdamDispatchResult dispatch_open_device(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	char* device;
	GdamSink* sink;
	int sink_id;

	/* Parse and insure the device name is there. */
	device = gdam_buffer_parse_string0(command);
	if (!device) {
		g_warning("error parsing device name");
		return GDAM_DISPATCH_INDICATE_FAIL;
	}

	/* Open the device. */
	sink = gdam_device_open(device);
	if (!sink) {
		gdam_debug("error opening device %s", device);
		g_free(device);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	
	if (gdam_dispatch_logging)
		gdam_dispatch_log("open_device \"%s\"", device);

	/* Obtain the sink id. */
	g_free(device);
	sink_id = gdam_dispatch_obtain_sink_id(sink);

	/* Respond to server. */
	gdam_buffer_append_int(rv, GDAM_RESPONSE_INT_SUCCESS);
	gdam_buffer_append_int(rv, sink_id);
	gdam_debug("open_device:  succeeded: id=%u", sink_id);
	return GDAM_DISPATCH_SUCCESS;
}
static GdamDispatchResult dispatch_feed_device(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id, sink_id;
	GdamSink* sink;
	GdamSource* source;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &sink_id)
	 || !gdam_buffer_parse_int(command, &source_id))
		return GDAM_DISPATCH_TOO_SHORT;

	source = gdam_namespace_local_source(source_id);
	sink = gdam_dispatch_get_sink(sink_id);
	if (!source || !sink) {
		if (sink)
			gtk_object_unref(GTK_OBJECT(sink));
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	if (gdam_dispatch_logging)
		gdam_dispatch_log("feed_device %d %d", sink_id, source_id);
	
	/* Feed the sound device. */
	gdam_sink_set_source(sink, source);
	gtk_object_unref(GTK_OBJECT(sink));
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}
static GdamDispatchResult dispatch_close_device(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int sink_id;
	GdamSink* sink;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &sink_id))
		return GDAM_DISPATCH_TOO_SHORT;

	/* Get the device. */
	sink = gdam_dispatch_get_sink(sink_id);
	if (!sink) 
		return GDAM_DISPATCH_INDICATE_FAIL;

	/* Destroy the device. */
	gtk_object_destroy(GTK_OBJECT(sink));
	gtk_object_unref(GTK_OBJECT(sink));

	if (gdam_dispatch_logging)
		gdam_dispatch_log("close_device %d", sink_id);

	/* Indicate success. */
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}

static gboolean has_initialized_hack = FALSE;
static gboolean has_hack = FALSE;

static GdamSource* gdam_dispatch_object_from_buffer_log(GdamBuffer* buffer,
                                                        gboolean    was_create)
{
	GdamSource* source;
	char* type;
	int obj_type;
	int num_args;
	GdamArg* args;
	if (!gdam_buffer_parse_int(buffer, &obj_type)) {
		gdam_debug("error parsing object type out of buffer");
		return NULL;
	}
	g_assert(obj_type == GDAM_PROTOCOL_SOURCE_ARG);
	if ((type=gdam_buffer_parse_string0(buffer)) == NULL) {
		gdam_debug("error parsing source type out of buffer");
		return NULL;
	}
	if (!has_initialized_hack) {
		if (getenv("GDAM_GPROF_HACK"))
			has_hack = TRUE;
		has_initialized_hack = TRUE;
	}

	if (has_hack && strcmp(type, "EXIT") == 0) {
		exit(0);
	}
	if (strncmp(type, "DEBUG",5) == 0) {
		gdam_server_debug = atoi(type + 5);
		g_message("DEBUG set to %d", gdam_server_debug);
		return NULL;
	}
	if (strcmp(type, "by_id") == 0) {
		int source_id;
		g_free(type);
		if (!gdam_buffer_parse_int(buffer, &source_id)) {
			gdam_debug("GDAM_SOURCE_BY_ID: parsing source id: "
				   "out of buffer");
			return NULL;
		}
		if (source_id == 0) {
			gdam_debug("GDAM_SOURCE_BY_ID: source_id was 0");
			return NULL;
		}
		source = gdam_namespace_local_source(source_id);
		if (!source) {
			gdam_debug("GDAM_SOURCE_BY_ID: no such source");
			return NULL;
		}
		return source;
	}
	if (!gdam_buffer_parse_int(buffer, &num_args)) {
		gdam_debug("error getting number of arguments for %s", type);
		g_free(type);
		return NULL;
	}
	args = gdam_arg_parse_from_buffer(buffer, num_args);
	source = gdam_source_new_generic_v(type, num_args, args);
	if (was_create && gdam_dispatch_logging) {
		char* argstring = gdam_arg_array_as_string(num_args, args);
		gdam_dispatch_log("source unnamed %s %s", type, argstring);
		g_free(argstring);
	}
	return source;
}
#if 0
GdamSource* gdam_dispatch_source_from_buffer(GdamBuffer* buffer)
{
	return gdam_dispatch_source_from_buffer_log(buffer, FALSE);
}
#endif

static GdamDispatchResult dispatch_create(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	GdamSource* source;
	int source_id;
	int local_source_id;
	source = gdam_dispatch_object_from_buffer_log(command, TRUE);
	if (!source) {
		gdam_debug("create_source: couldn't parse buffer");
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	source_id = source->source_id;

	/* Also, allocate a "local" source id -- just local to this
	 * request. */
	local_source_id = gdam_namespace_alloc_local_source(source);
	gdam_debug("dispatch_create_source: created source id %d, "
		   "local source id %d", source_id, local_source_id);
	gdam_buffer_append_int(rv,
		GDAM_RESPONSE_INT_SUCCESS);
	gdam_buffer_append_int(rv, source_id);
	return GDAM_DISPATCH_SUCCESS;
}
typedef GtkObject*   (*LookupFunc)(guint id);

static GtkObject* lookup_source(guint id)
{
	GdamSource* source = gdam_namespace_local_source(id);
	GtkObject* rv;
	if (source == NULL)
		return NULL;
	rv = GTK_OBJECT(source);
	gtk_object_ref(rv);
	return rv;
}

static GtkObject* lookup_buffer(guint id)
{
	GtkObject* rv;
	GdamCacheBuffer*  cache_buffer = gdam_cache_buffer_get_by_id(id);
	if (cache_buffer == NULL)
		return NULL;
	rv = GTK_OBJECT(cache_buffer);
	gtk_object_ref(rv);
	return rv;
}

static LookupFunc get_lookup_func_for_arena(guint arena_id)
{
	switch (arena_id) {
		case GDAM_PROTOCOL_SOURCE_ARG:
			return lookup_source;
		case GDAM_PROTOCOL_BUFFER_ARG:
			return lookup_buffer;
	}
	gdam_debug("bad arena-id to set or get args");
	return NULL;
}
		
static GdamDispatchResult dispatch_source_set_args(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id, num_args;
	GtkObject* object;
	GdamArg* args;
	int i;
	LookupFunc lookup_func;

	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id)
	 || !gdam_buffer_parse_int(command, &num_args))
	 	return GDAM_DISPATCH_TOO_SHORT;
	if (source_id == 0)
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	if (num_args == 0)
	 	return GDAM_DISPATCH_INDICATE_SUCCESS;
	
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no source of id %d to set args", source_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}

	args = gdam_arg_parse_from_buffer(command, num_args);
	if (!args) {
		gdam_debug("error parsing arguments");
		gtk_object_unref(object);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}

	if (gdam_dispatch_logging) {
		char* argstring = gdam_arg_array_as_string(num_args, args);
		gdam_dispatch_log("set %d %s", source_id, argstring);
		g_free(argstring);
	}

	{
		GdamConnection* last_con = gdam_connection_for_arg;
		gdam_connection_for_arg = connection;
		/* TODO: there should be an array set maybe... */
		for (i = 0; i < num_args; i++)
			gdam_arg_set_object_arg(object, args + i);
		gdam_connection_for_arg = last_con;
	}

	gdam_arg_free_array(num_args, args);
	gtk_object_unref(object);
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}

static GdamDispatchResult dispatch_get_args(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id, num_args;
	int i;
	char** args;
	GtkObject* object;
	GdamArg* rv_args;
	LookupFunc lookup_func;
	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id)
	 || !gdam_buffer_parse_int(command, &num_args))
	 	return GDAM_DISPATCH_TOO_SHORT;
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no object of id %d to get args", source_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	args = (char**)alloca(sizeof(char*) * num_args);
	for (i = 0; i < num_args; i++) {
		if ((args[i] = gdam_buffer_parse_string0(command)) == NULL) {
			gdam_debug("too few arguments to get_arg command");
			{ int j; for (j = 0; j < i; j++) g_free(args[i]); }
			gtk_object_unref(object);
			return GDAM_DISPATCH_TOO_SHORT;
		}
	}

	{
		GdamConnection* last_con = gdam_connection_for_arg;
		gdam_connection_for_arg = connection;
		gdam_arg_get_by_name(object,
				     num_args,
				     (const char* const*) args,
				     &num_args,
				     &rv_args);
		gdam_connection_for_arg = last_con;
	}

	gdam_buffer_append_int(rv,
		GDAM_RESPONSE_ARG_SUCCESS);
	gdam_buffer_append_int(rv, num_args);
	gdam_arg_write_to_buffer(rv, num_args, rv_args);
	gdam_arg_free_array(num_args, rv_args);

	gtk_object_unref(object);
	return GDAM_DISPATCH_SUCCESS;;
}

static GdamDispatchResult dispatch_get_all_args(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id, only_vital;
	int num_args;
	LookupFunc lookup_func;
	GdamArg* rv_args;
	GtkObject* object;
	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id)
	 || !gdam_buffer_parse_int(command, &only_vital))
	 	return GDAM_DISPATCH_TOO_SHORT;
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no object of id %d to get all args", source_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
		
	{
		GdamConnection* last_con = gdam_connection_for_arg;
		gdam_connection_for_arg = connection;
		gdam_arg_get_all(object,
				 only_vital,
				 &num_args,
				 &rv_args);
		gdam_connection_for_arg = last_con;
	}

	gdam_buffer_append_int(rv,
		GDAM_RESPONSE_ARG_SUCCESS);
	gdam_buffer_append_int(rv, num_args);
	gdam_arg_write_to_buffer(rv, num_args, rv_args);
	gdam_arg_free_array(num_args, rv_args);

	return GDAM_DISPATCH_SUCCESS;
}

static GdamDispatchResult dispatch_call_function(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id, num_args;
	GtkObject* object;
	GdamArg* args;
	int num_rv_args;
	GdamArg* rv_args;
	char* function_name = NULL;
	LookupFunc lookup_func;

	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id)
	 || !(function_name = gdam_buffer_parse_string0(command))
	 || !gdam_buffer_parse_int(command, &num_args)) {
		g_free(function_name);
	 	return GDAM_DISPATCH_TOO_SHORT;
	}
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
		g_free(function_name);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no source of id %d to set args", source_id);
		g_free(function_name);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}


	args = gdam_arg_parse_from_buffer(command, num_args);
	if (!args) {
		gdam_debug("error parsing arguments");
		gtk_object_unref(object);
		g_free(function_name);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}

	if (gdam_dispatch_logging) {
		char* argstring = gdam_arg_array_as_string(num_args, args);
		gdam_dispatch_log("call %d %s", source_id, argstring);
		g_free(argstring);
	}
	{
		GdamConnection* last_con = gdam_connection_for_arg;
		gdam_connection_for_arg = connection;
		if (!gdam_arg_call_function(object,
					    function_name,
					    num_args,
					    args,
					    &num_rv_args,
					    &rv_args)) {
			gdam_debug("error calling function %s", function_name);
			gtk_object_unref(object);
			gdam_connection_for_arg = last_con;
			return GDAM_DISPATCH_INDICATE_FAIL;
		}
		gdam_connection_for_arg = last_con;
	}

	gdam_buffer_append_int(rv, GDAM_RESPONSE_ARG_SUCCESS);
	gdam_buffer_append_int(rv, num_rv_args);
	gdam_arg_write_to_buffer(rv, num_rv_args, rv_args);
	gdam_arg_free_array(num_rv_args, rv_args);
	gdam_arg_free_array(num_args, args);

	return GDAM_DISPATCH_SUCCESS;
}


static GdamDispatchResult dispatch_configure_source(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id;
	GdamSource* source;
	GdamBuffer* returned_buffer;
	gboolean success;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &source_id))
		return GDAM_DISPATCH_TOO_SHORT;

	source = gdam_namespace_local_source(source_id);
	if (!source) {
		gdam_debug("source %d not found", source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	returned_buffer = gdam_source_interpret_command(source,
	                                                command,
	                                                &success);
	if (!returned_buffer)
		return success ? GDAM_DISPATCH_INDICATE_SUCCESS
		               : GDAM_DISPATCH_INDICATE_FAIL;
	gdam_buffer_append_buffer(rv, returned_buffer, TRUE);
	return success ? GDAM_DISPATCH_SUCCESS : GDAM_DISPATCH_FAIL;
}

#if 0		/* not supported generically yet. */
static GdamDispatchResult dispatch_attach_source(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id, sink_id;
	GdamSource* sink;
	GdamSource* source;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &sink_id)
	 || !gdam_buffer_parse_int(command, &source_id))
		return GDAM_DISPATCH_TOO_SHORT;

	source = gdam_namespace_local_source(source_id);
	/* our sink is another source, so the following is not a typo: */
	sink = gdam_namespace_local_source(sink_id);
	if (!source || !sink) {
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	
	/* Feed the sink source. */
	gdam_sink_set_source(sink, source);
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}
#endif
static GdamDispatchResult dispatch_detach_source(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id;
	GdamSource* source;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &source_id))
		return GDAM_DISPATCH_TOO_SHORT;

	source = gdam_namespace_local_source(source_id);
	if (!source)
		return GDAM_DISPATCH_INDICATE_FAIL;

	gdam_debug("gdam_source_detach: source_id %d", source_id);
	gdam_source_orphan(source);

	/* orphan refs it once */
	gtk_object_unref((GtkObject*)source);

	return GDAM_DISPATCH_INDICATE_SUCCESS;
}
static GdamDispatchResult dispatch_synchronize(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	GdamBuffer subs;

	int mobile_source_id, fixed_source_id;
	GdamTime mobile_beat_period, mobile_start_beat;
	GdamTime fixed_beat_period, fixed_start_beat;
	GdamSource* mobile_source;
	GdamSource* fixed_source;
	int to_seek;

	gdam_buffer_construct(&subs);
	if (!gdam_buffer_parse_int(command, &mobile_source_id)
	 || !gdam_buffer_parse_int(command, &fixed_source_id)
	 || !gdam_buffer_parse_time(command, &mobile_beat_period)
	 || !gdam_buffer_parse_time(command, &mobile_start_beat)
	 || !gdam_buffer_parse_time(command, &fixed_beat_period)
	 || !gdam_buffer_parse_time(command, &fixed_start_beat))
		return GDAM_DISPATCH_INDICATE_FAIL;

	mobile_source = gdam_namespace_local_source(mobile_source_id);
	if (!mobile_source) {
		gdam_debug("synchronize: error finding mobile source of id %d",
				mobile_source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	fixed_source = gdam_namespace_local_source(fixed_source_id);
	if (!fixed_source) {
		gdam_debug("synchronize: error finding fixed source of id %d",
				fixed_source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	to_seek = gdam_source_get_sync_seek(mobile_source, fixed_source,
					    mobile_beat_period,
					    mobile_start_beat,
					    fixed_beat_period,
					    fixed_start_beat,
					    NULL);
	if (to_seek != 0)
		gdam_source_seek(mobile_source, to_seek, 1);
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}
static GdamDispatchResult dispatch_sync_preview(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	GdamBuffer subs;

	int mobile_source_id, fixed_source_id;
	GdamTime mobile_beat_period, mobile_start_beat;
	GdamTime fixed_beat_period, fixed_start_beat;
	GdamSource* mobile_source;
	GdamSource* fixed_source;
	int to_seek;

	gdam_buffer_construct(&subs);
	if (!gdam_buffer_parse_int(command, &mobile_source_id)
	 || !gdam_buffer_parse_int(command, &fixed_source_id)
	 || !gdam_buffer_parse_time(command, &mobile_beat_period)
	 || !gdam_buffer_parse_time(command, &mobile_start_beat)
	 || !gdam_buffer_parse_time(command, &fixed_beat_period)
	 || !gdam_buffer_parse_time(command, &fixed_start_beat))
		return GDAM_DISPATCH_INDICATE_FAIL;

	mobile_source = gdam_namespace_local_source(mobile_source_id);
	if (!mobile_source) {
		gdam_debug("synchronize: error finding mobile source of id %d",
				mobile_source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	fixed_source = gdam_namespace_local_source(fixed_source_id);
	if (!fixed_source) {
		gdam_debug("synchronize: error finding fixed source of id %d",
				fixed_source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	to_seek = gdam_source_get_sync_seek(mobile_source, fixed_source,
					    mobile_beat_period,
					    mobile_start_beat,
					    fixed_beat_period,
					    fixed_start_beat,
					    NULL);
	gdam_buffer_append_int(rv, GDAM_RESPONSE_TIME_SUCCESS);
	gdam_buffer_append_time(rv, to_seek);
	return GDAM_DISPATCH_SUCCESS;
}
static GdamDispatchResult dispatch_reopen_device(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int device_id;
	GdamSink* sink;
	if (!gdam_buffer_parse_int(command, &device_id))
		return GDAM_DISPATCH_INDICATE_FAIL;
	sink = gdam_dispatch_get_sink(device_id);
	if (sink == NULL) {
		gdam_debug("device %d wasn't open", device_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	if (sink->device_name == NULL) {
		gdam_debug("device %d wasn't in config file", device_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	gdam_device_reopen(sink->device_name);
	return GDAM_DISPATCH_INDICATE_SUCCESS;
	
}
static GdamDispatchResult dispatch_package(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	GdamBuffer subs;
	GdamBuffer* sub;
	int num_subs;
	int i;
	gdam_buffer_construct(&subs);
	if (!gdam_buffer_parse_int(command, &num_subs))
		return GDAM_DISPATCH_INDICATE_FAIL;

	if (gdam_dispatch_logging) {
		gdam_dispatch_log("begin_package");
		dispatch_logging_indent++;
	}

	gdam_source_defer_done_constructing();
	for (i = 0; i < num_subs; i++) {
		GdamBuffer tmp;
		int length;
		gdam_buffer_construct(&tmp);
		if (!gdam_buffer_peek_int(command, &length)) {
			gdam_source_undefer_done_constructing();
			return GDAM_DISPATCH_TOO_SHORT;
		}
		gdam_buffer_send_buffer(&tmp, command, length, TRUE);
		gdam_buffer_parse_int(&tmp, &length);
		sub = gdam_dispatch_buffer(connection, perms, &tmp);
		gdam_buffer_destruct(&tmp);
		if (!sub) {
			gdam_source_undefer_done_constructing();
			gdam_buffer_destruct(&subs);
			return GDAM_DISPATCH_INDICATE_FAIL;
		}
		gdam_buffer_prepend_length(sub);
		gdam_buffer_append_buffer(&subs, sub, TRUE);
		gdam_buffer_destroy(sub);
	}
	gdam_source_undefer_done_constructing();
	if (gdam_dispatch_logging) {
		dispatch_logging_indent--;
		gdam_dispatch_log("end_package");
	}
	gdam_buffer_append_int(rv, GDAM_RESPONSE_PACKAGE);
	gdam_buffer_append_int(rv, num_subs);
	gdam_buffer_append_buffer(rv, &subs, TRUE);
	return GDAM_DISPATCH_SUCCESS;
}

static GdamDispatchResult dispatch_get_named_buffers
                                               (GdamConnection*  connection,
		                                GdamPermissions* perms,
		                                GdamBuffer*      command,
		                                GdamBuffer*      rv)
{
	char** common;
	int num_buffer;
	gdam_wavetable_init_common();
	common = gdam_wavetable_get_list_of_common();
	gdam_buffer_append_int(rv, GDAM_RESPONSE_STRING_ARRAY_SUCCESS);
	for (num_buffer = 0; common[num_buffer] != NULL; num_buffer++)
		;
	gdam_buffer_append_int(rv, num_buffer);
	for (num_buffer = 0; common[num_buffer] != NULL; num_buffer++)
		gdam_buffer_append_string0(rv, common[num_buffer]);
	g_free(common);
	return GDAM_DISPATCH_SUCCESS;
}

#if 0
static GdamDispatchResult dispatch_get_sound_spec(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id, recurse;
	GdamSource* to_write;

	if (!gdam_buffer_parse_int(command, &source_id) 
         || !gdam_buffer_parse_int(command, &recurse)) {
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	to_write = gdam_namespace_local_source(source_id);
	if (!to_write) {
		gdam_debug("dispatch_get_sound_spec: no source with id %d",
				source_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	gdam_buffer_append_int(rv, GDAM_RESPONSE_DATA_SUCCESS);
	{
		GdamBuffer tmp;
		gdam_buffer_construct(&tmp);
		gdam_source_write_to_binary(to_write,
		                      &tmp,
				      recurse ? 0 : 
					GDAM_SOURCE_WRITE_NO_RECURSE);
		gdam_buffer_append_int(rv, tmp.size);
		gdam_buffer_append_buffer(rv, &tmp, TRUE);
	}

	return GDAM_DISPATCH_SUCCESS;
}
#endif
static GdamDispatchResult dispatch_start_caching(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	GdamSource* source;
	guint buffer_id, event_con_id;
	int source_id;
	GdamCacheBuffer* cache;

	if (!gdam_buffer_parse_int(command, &source_id)) {
		g_warning("error parsing source_id");
		return GDAM_DISPATCH_TOO_SHORT;
	}
	source = gdam_namespace_local_source(source_id);
	if (!source) {
		g_warning("error parsing sound spec to cache");
		return GDAM_DISPATCH_INDICATE_FAIL;
	}

	cache = gdam_cache_buffer_new(source);
	buffer_id = cache->buffer_id;

	gdam_connection_ref(connection, GTK_OBJECT(cache));

	event_con_id = gdam_event_connect(GTK_OBJECT(cache),
	                                  connection,
                                          GDAM_CACHE_BUFFER_EVENT_DONE,
					  NULL);

	gdam_buffer_append_int(rv, GDAM_RESPONSE_CACHING_STARTED);
	gdam_buffer_append_int(rv, buffer_id);
	gdam_buffer_append_int(rv, event_con_id);
	return GDAM_DISPATCH_SUCCESS;
}

#if 0
static GdamDispatchResult dispatch_uncache(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	guint buffer_id;
	if (!gdam_buffer_parse_int(command, &buffer_id))
		return GDAM_DISPATCH_TOO_SHORT;

	return gdam_cache_buffer_unregister(buffer_id)
				? GDAM_DISPATCH_INDICATE_SUCCESS
				: GDAM_DISPATCH_INDICATE_FAIL;
}
#endif
static GdamDispatchResult dispatch_get_cache_profile(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int buffer_id, mode, sample_start, sample_length, num_outputs, i;
	gint* returned_buffer;
	gint32* tmp;
	GdamCacheBuffer* cache_buffer;
	if (!gdam_buffer_parse_int(command, &buffer_id)
	 || !gdam_buffer_parse_int(command, &mode)
	 || !gdam_buffer_parse_int(command, &sample_start)
	 || !gdam_buffer_parse_int(command, &sample_length)
	 || !gdam_buffer_parse_int(command, &num_outputs)) {
	 	return GDAM_DISPATCH_TOO_SHORT;
	}
	cache_buffer = gdam_cache_buffer_get_by_id(buffer_id);
	if (cache_buffer == NULL) {
		gdam_debug("no cache buffer of id %d to profile", buffer_id);
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	returned_buffer = gdam_cache_buffer_profile(cache_buffer,
				                    mode,
                                                    sample_start,
				                    sample_length,
				                    num_outputs);
	if (returned_buffer == NULL) {
		gdam_debug("error getting cache buffer");
		return GDAM_DISPATCH_INDICATE_FAIL;
	}

	
	gdam_buffer_append_int(rv, GDAM_RESPONSE_DATA_SUCCESS);
	gdam_buffer_append_int(rv, num_outputs);

	if (mode == GDAM_PROTOCOL_PROFILE_STEREO)
		num_outputs *= 2;

	tmp = alloca(sizeof(gint32) * num_outputs);

	for (i = 0; i < num_outputs; i++)
		tmp[i] = GINT32_TO_BE(returned_buffer[i]);

	g_free(returned_buffer);
	gdam_buffer_append(rv, tmp, sizeof(gint32) * num_outputs);
	return GDAM_DISPATCH_SUCCESS;
}

static GdamDispatchResult dispatch_get_server_side_int(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	char* ns;
	char* name;
	int val;

	/* Parse and error check. */
	ns = gdam_buffer_parse_string0(command);
	name = gdam_buffer_parse_string0(command);
	if (!name) {
		g_free(ns);
		g_warning("error parsing device name");
		return GDAM_DISPATCH_INDICATE_FAIL;
	}

	if (gdam_namespace_int_lookup(ns, name, &val)) {
		gdam_buffer_append_int(rv, GDAM_RESPONSE_INT_SUCCESS);
		gdam_buffer_append_int(rv, val);
		return GDAM_DISPATCH_SUCCESS;
	}
	return GDAM_DISPATCH_INDICATE_FAIL;
}
#if 0
static GdamDispatchResult dispatch_feed_device(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id, sink_id;
	GdamSink* sink;
	GdamSource* source;

	/* Parse, as usual. */
	if (!gdam_buffer_parse_int(command, &sink_id)
	 || !gdam_buffer_parse_int(command, &source_id))
		return GDAM_DISPATCH_TOO_SHORT;

	source = gdam_namespace_local_source(source_id);
	sink = gdam_namespace_local_source(sink_id);
	if (!source || !sink) {
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	
	/* Feed the sound device. */
	gdam_sink_set_source(sink, source);
	return GDAM_DISPATCH_INDICATE_SUCCESS;
}
#endif
static GdamDispatchResult dispatch_subscribe_source(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int source_id, subscription_type;
	guint connection_id = 0;
	GdamSource* source;
	if (!gdam_buffer_parse_int(command, &source_id)
	 || !gdam_buffer_parse_int(command, &subscription_type))
		return GDAM_DISPATCH_TOO_SHORT;
	source = gdam_namespace_local_source(source_id);
	if (!source)
		return GDAM_DISPATCH_INDICATE_FAIL;
	if (subscription_type == GDAM_SUBSCRIPTION_ENDED) {
		connection_id = gdam_event_connect(GTK_OBJECT(source),
						   connection,
				                   GDAM_SOURCE_EVENT_ENDED,
				                   NULL);
	} else if (subscription_type == GDAM_SUBSCRIPTION_TIMER) {
		int period = 500;
		if (!gdam_buffer_parse_int(command, &period))
			g_warning("error parsing period for source timer");
		connection_id = gdam_timer_connect(
					GTK_OBJECT(source),
					GDAM_SOURCE_EVENT_TIMER,
					connection,
					period,
					NULL);
	} else {
		gdam_debug("unknown subscription type %d", subscription_type);
	}
	if (connection_id == 0) {
		return GDAM_DISPATCH_INDICATE_FAIL;
	}
	gdam_buffer_append_int(rv, GDAM_RESPONSE_INT_SUCCESS);
	gdam_buffer_append_int(rv, connection_id);
	return GDAM_DISPATCH_SUCCESS;
}

static GdamDispatchResult dispatch_unsubscribe(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int subscription_id;
	if (!gdam_buffer_parse_int(command, &subscription_id))
		return GDAM_DISPATCH_TOO_SHORT;
	if (gdam_event_disconnect(subscription_id))
		return GDAM_DISPATCH_INDICATE_SUCCESS;
	return GDAM_DISPATCH_INDICATE_FAIL;
}
static GdamDispatchResult dispatch_ref(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id;
	LookupFunc lookup_func;
	GtkObject* object;
	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id))
	 	return GDAM_DISPATCH_TOO_SHORT;
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no object of id %d to ref", source_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	if (gdam_connection_ref(connection, object)) {
		gtk_object_unref(object);
		return GDAM_DISPATCH_INDICATE_SUCCESS;
	}
	gtk_object_unref(object);
	return GDAM_DISPATCH_INDICATE_FAIL;
}

static GdamDispatchResult dispatch_unref(
		GdamConnection* connection,
		GdamPermissions* perms,
		GdamBuffer* command,
		GdamBuffer* rv)
{
	int arena_id, source_id;
	LookupFunc lookup_func;
	GtkObject* object;
	if (!gdam_buffer_parse_int(command, &arena_id)
	 || !gdam_buffer_parse_int(command, &source_id))
	 	return GDAM_DISPATCH_TOO_SHORT;
	lookup_func = get_lookup_func_for_arena(arena_id);
	if (lookup_func == NULL) {
		gdam_debug("no arena for id %d", arena_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	object = (*lookup_func)(source_id);
	if (!object) {
		gdam_debug("no object of id %d to ref", source_id);
	 	return GDAM_DISPATCH_INDICATE_FAIL;
	}
	gtk_object_unref(object);
	if (gdam_connection_unref(connection, object))
		return GDAM_DISPATCH_INDICATE_SUCCESS;
	return GDAM_DISPATCH_INDICATE_FAIL;
}

/*
 * Initialize the builtin method table.
 *
 * Perhaps an array would be better, but this has its
 * advantages...
 */

struct {
	int method;
	GdamDispatchMessageFunc dispatcher;
} builtin_methods[] = {
	{ GDAM_MESSAGE_OPEN_DEVICE, dispatch_open_device },
	{ GDAM_MESSAGE_FEED_DEVICE, dispatch_feed_device },
	{ GDAM_MESSAGE_CLOSE_DEVICE, dispatch_close_device },
	{ GDAM_MESSAGE_CREATE, dispatch_create },
	/* { GDAM_MESSAGE_ATTACH_SOURCE, dispatch_attach_source }, */
	{ GDAM_MESSAGE_CONFIGURE_SOURCE, dispatch_configure_source },
	{ GDAM_MESSAGE_DETACH_SOURCE, dispatch_detach_source },
	{ GDAM_MESSAGE_GET_CACHE_PROFILE, dispatch_get_cache_profile },
	{ GDAM_MESSAGE_SET_ARGS, dispatch_source_set_args },
	{ GDAM_MESSAGE_GET_ARGS, dispatch_get_args },
	{ GDAM_MESSAGE_GET_ALL_ARGS, dispatch_get_all_args },
	{ GDAM_MESSAGE_CALL_FUNCTION, dispatch_call_function },
	{ GDAM_MESSAGE_START_CACHING, dispatch_start_caching },
	//{ GDAM_MESSAGE_UNCACHE, dispatch_uncache },
	{ GDAM_MESSAGE_GET_SERVER_INT, dispatch_get_server_side_int },
	{ GDAM_MESSAGE_SYNCHRONIZE, dispatch_synchronize },
	{ GDAM_MESSAGE_SYNC_PREVIEW, dispatch_sync_preview },
	{ GDAM_MESSAGE_REOPEN_DEVICE, dispatch_reopen_device },
	{ GDAM_MESSAGE_PACKAGE, dispatch_package },
	{ GDAM_MESSAGE_SUBSCRIBE_SOURCE, dispatch_subscribe_source },
	{ GDAM_MESSAGE_UNSUBSCRIBE, dispatch_unsubscribe },
	{ GDAM_MESSAGE_GET_NAMED_BUFFERS, dispatch_get_named_buffers },
	{ GDAM_MESSAGE_REF, dispatch_ref },
	{ GDAM_MESSAGE_UNREF, dispatch_unref }
};
#ifdef DIM
#undef DIM
#endif
#define DIM(arr)      ((int)(sizeof(arr) / sizeof(arr[0])))

GHashTable* gdam_dispatch_methods = NULL;

static GHashTable* make_builtin_methods()
{
	GHashTable* rv = g_hash_table_new(NULL, NULL);
	int i;
	for (i = 0; i < DIM(builtin_methods); i++)
		g_hash_table_insert(
			rv,
			GINT_TO_POINTER(builtin_methods[i].method),
			(gpointer) (builtin_methods[i].dispatcher)
		);
	return rv;
}
static inline void initialize_builtin_method_hash()
{
	if (!gdam_dispatch_methods)
		gdam_dispatch_methods = make_builtin_methods();
}

GdamDispatchMessageFunc gdam_dispatch_message_find_by_type(int i)
{
	gpointer rv;
	initialize_builtin_method_hash();
	rv = g_hash_table_lookup(gdam_dispatch_methods, GINT_TO_POINTER(i));
	return (GdamDispatchMessageFunc) rv;
}

