/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamtee.h"

/*
 * Strategy:
 *
 *    Tees     hold a reference to their TeeSink,
 *             and trap its destroy signal.
 *             Each Tee stores the TeeSink-relative
 *             timing data.
 *    TeeSinks just store timing information and
 *             coordinate the global reads.
 *
 *  - no reading is done unless it is read from
 *    by the first tee source (via someone calling
 *    its read() method).
 *  - when attaching to an existing sink, GdamTees will
 *    be reading at the latest position.
 *  - when a tee is first created it has no sink.
 *
 * The sink is kept entirely private except having ids
 * as far as clients are concerned.
 *
 * Here are the various arguments to a GdamTee:
 *    
 *     sink_id         gets/sets the unique identifier
 *                     of the tee sink.
 *                     When a source is first created,
 *                     its sink_id is 0, but you may not 
 *                     set the sink_id to 0.
 *
 *     create_sink     makes a new tee sink for this source;
 *                     the sink id is returned.
 *
 *     set_source_id   sets the tee_sinks underlying source,
 *                     making a new tee sink if necessary.
 */

static void     gdam_tee_sink_destroy    (GtkObject*  object);
static void     gdam_tee_destroy         (GtkObject*  object);
static int      gdam_tee_read            (GdamSource* source,
                                          GdamSample* buf,
				          int         num);
static gboolean gdam_tee_is_ready_to_read(GdamSource* source);
static void     gdam_tee_miss_data       (GdamSource* source,
                                          GdamSample* buf,
				          int         num);

/*
 *   ____     _               _____         ____  _       _
 *  / ___| __| | __ _ _ __ __|_   _|__  ___/ ___|(_)_ __ | | __
 * | |  _ / _` |/ _` | '_ ` _ \| |/ _ \/ _ \___ \| | '_ \| |/ /
 * | |_| | (_| | (_| | | | | | | |  __/  __/___) | | | | |   <
 *  \____|\__,_|\__,_|_| |_| |_|_|\___|\___|____/|_|_| |_|_|\_\
 * 
 */
#define GDAM_TEE_DEFAULT_BUFFER_MAX	(1<<16)

static GHashTable* tee_sinks_by_id = NULL;

static GtkObjectClass* parent_class = NULL;

static GdamTeeSink* gdam_tee_sink_get_by_id(guint id)
{
	if (tee_sinks_by_id == NULL)
		return NULL;
	return g_hash_table_new(tee_sinks_by_id, GUINT_TO_POINTER(id));
}

static int get_unique_tee_sink_id()
{
	static int id = 0;
	return ++id;
}

static void gdam_tee_sink_init(GdamTeeSink* sink)
{
	sink->buffer_max = GDAM_TEE_DEFAULT_BUFFER_MAX;
	sink->bytes_sunk = 0;
	sink->id = get_unique_tee_sink_id();
	gdam_buffer_construct(&sink->buffer);
	if (tee_sinks_by_id == NULL)
		tee_sinks_by_id = g_hash_table_new(NULL, NULL);
	g_hash_table_insert(tee_sinks_by_id,
	                    GINT_TO_POINTER(sink->id),
			    sink);
}

static void gdam_tee_sink_class_init(GdamSinkClass* sink_class)
{
	
	GtkObjectClass* object_class = (GtkObjectClass*)sink_class;
	object_class->destroy = gdam_tee_sink_destroy;
	parent_class = gtk_type_class(GDAM_TYPE_SINK);
}

static void gdam_tee_sink_new()
{
	return GDAM_TEE_SINK(gtk_type_new(GDAM_TYPE_TEE_SINK));
}

GtkType gdam_tee_sink_get_type()
{
	static GtkType tee_sink_type = 0;
	if (!tee_sink_type) {
		static const GtkTypeInfo tee_sink_info = {
			"GdamTeeSink",
			sizeof(GdamTeeSink),
			sizeof(GdamTeeSinkClass),
			(GtkClassInitFunc) gdam_tee_sink_class_init,
			(GtkObjectInitFunc) gdam_tee_sink_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		tee_sink_type = gtk_type_unique(GTK_TYPE_OBJECT,
		                                &tee_sink_info);
	}
	return tee_sink_type;
}

static void gdam_tee_sink_destroy(GtkObject* object)
{
	GdamTeeSink* tee_sink = GDAM_TEE_SINK(object);
	gdam_buffer_destruct(&tee_sink->buffer);
	(*parent_class->destroy)(object);
}


/*
 *   ____     _               _____
 *  / ___| __| | __ _ _ __ __|_   _|__  ___
 * | |  _ / _` |/ _` | '_ ` _ \| |/ _ \/ _ \
 * | |_| | (_| | (_| | | | | | | |  __/  __/
 *  \____|\__,_|\__,_|_| |_| |_|_|\___|\___|
 * 
 */

static void gdam_tee_destroy(GtkObject* object)
{
	GdamTee* tee = (GdamTee*)object;
	if (tee->tee_sink) {
		GtkObject* to_unref = GTK_OBJECT(tee->tee_sink);
		tee->tee_sink = NULL;
		gtk_object_unref(to_unref);
	}
}

/*=====================================================================*
 *                     gdam_tee_sink_append_samples                    *
 *                                                                     *
 * Append raw samples to a tee sink's buffer.                          *
 * Only used internally, of course...                                  *
 *                                                                     *
 * The point is that gdam_tee_sink_append_samples                      *
 * and gdam_tee_sink_append_zeroes should be the                       *
 * only way to change a GdamTeeSink.  (every change goes               *
 * though those interfaces)                                            *
 *                                                                     *
 * Parameters:                                                         *
 *     sink              the tee_sink whose buffers should have        *
 *                       samples appended                              *
 *     samples           the sample buffer.                            *
 *     num_read          number of samples to append                   *
 *=====================================================================*/
static void gdam_tee_sink_append_samples(GdamTeeSink* sink,
                                         GdamSample*  samples,
					 int          num_samples)
{
	sink->samples_read += num_samples;
	sink->num_buffered += num_samples;
	if (sink->num_buffered > sink->buffer_size)
		sink->num_buffered = sink->buffer_size;

	if (sink->buffer_max < num_samples) {
		int too_much = sink->buffer_max - num_samples;
		int num_discarded;
		samples += too_much;
		num_samples -= too_much;	/* == sink->buffer_max */
	}

	/* Now, append the data to a circular buffer. */

	/* If the buffer needs to be divided, this takes
	 * care of the first part. */
	if (num_samples > sink->buffer_size - sink->read_index) {
		int num_first_part = sink->buffer_size - sink->read_index;
		gdam_sample_copy(sink->buffer + sink->read_index,
		                 samples,
				 num_first_part);
		samples += num_first_part;
		num_samples -= num_first_part;
		sink->read_index = 0;
	}

	g_assert (num_samples <= sink->buffer_size - sink->read_index);

	/* Second or only part of circular buffer. */
	gdam_sample_copy(sink->buffer + sink->read_index,
	                 samples,
			 num_samples);
	sink->read_index += num_samples;
	if (sink->read_index >= sink->buffer_size)
		sink->read_index -= sink->buffer_size;

	g_assert (sink->read_index < sink->buffer_size);
}

static void gdam_tee_sink_append_zeroes (GdamTeeSink* sink,
					 int          num_zeroes)
{
	sink->samples_read += num_samples;
	sink->num_buffered += num_samples;
	if (sink->num_buffered > sink->buffer_size)
		sink->num_buffered = sink->buffer_size;
	if (sink->buffer_max < num_zeroes)
		num_samples = sink->buffer_max;

	/* Now, append the data to a circular buffer. */

	/* If the buffer needs to be divided, this takes
	 * care of the first part. */
	if (num_zeroes > sink->buffer_size - sink->read_index) {
		int num_first_part = sink->buffer_size - sink->read_index;
		gdam_sample_zero(sink->buffer + sink->read_index,
				 num_first_part);
		num_zeroes -= num_first_part;
		sink->read_index = 0;
	}

	g_assert (num_zeroes <= sink->buffer_size - sink->read_index);

	/* Second or only part of circular buffer. */
	gdam_sample_zero(sink->buffer + sink->read_index, num_zeroes);
	sink->read_index += num_zeroes;
	if (sink->read_index >= sink->buffer_size)
		sink->read_index -= sink->buffer_size;

	g_assert (sink->read_index < sink->buffer_size);
}

/*
 *                     gdam_tee_sink_read_try_read_to
 *
 * Parameters:
 *     sink       the tee_sink to tried to ready.
 *     desired    the position (in total number of samples output) to
 *                try to seek to.
 */
static void gdam_tee_sink_read_try_read_to(GdamTeeSink* sink, GdamTime desired)
{
	int num_desired;
	GdamSample* buf;
	GdamSource* feed;

	feed = GDAM_SINK(sink)->feed;

	if (feed == NULL)
		return;

	/* This block just makes num_desired = desired - sink->samples_read
	 * 64-bit safe. */
	{
		gint64 tmp_num_desired = desired - sink->samples_read;
		if (tmp_num_desired < 0 || tmp_num_desired > TOO_MANY_TO_READ)
			return;
		num_desired = (int) tmp_num_desired;
	}

	buf = (GdamSample*) alloca(sizeof(GdamSample) * num_desired);
	num_read = gdam_source_read(feed, buf, num_desired);

	gdam_tee_sink_append_samples(sink, buf, num_read);
}

static int gdam_tee_read(GdamSource* source,
                         GdamSample* buf,
			 int         num)
{
	GdamTee* tee = GDAM_TEE(source);
	GdamTeeSink* tee_sink = tee->tee_sink;
	GdamSource* feed;
	if (tee_sink == NULL)
		return NULL;
	feed = GDAM_SINK(tee_sink)->feed;

	if (tee->tee_time <= tee_sink->samples_read + num) {
		if (tee->tee_time < tee_sink->samples_read) {
			/* This print is pretty stupid.  Hopefully
			 * it won't alarm anyone. (shouldn't happen) */
			g_warning("gdam_tee_read: hmmm... "
			          "we've been read from even though we "
			          "aren't caught up.");
		}
		gdam_tee_sink_read_try_read_to(tee->tee_sink,
		                               tee_sink->samples_read + num);
	}

	if (tee->tee_time <= tee_sink->samples_read) {
		/* No actual data to return, but we've tried. */
		...
	}
	
	if (feed == NULL) {
		...
	}
	...
}

static gboolean gdam_tee_is_ready_to_read(GdamSource* source)
{
	GdamTee* tee = GDAM_TEE(source);
	GdamTeeSink* tee_sink = tee->tee_sink;
	if (tee_sink == NULL)
		return TRUE;
	...
}

static void gdam_tee_miss_data(GdamSource* source,
                               GdamSample* buf,
			       int         num)
{
}

static void gdam_tee_init(GdamTee* sink)
{
}

static void gdam_tee_arg_set_tee_sink(GtkObject* object, GdamArg* arg)
{
	GdamTee* tee = GDAM_TEE(object);
	...
}

static void gdam_tee_arg_get_tee_sink(GtkObject* object, GdamArg* arg)
{
	GdamTee* tee = GDAM_TEE(object);
	GDAM_ARG_INT(arg) = (tee->tee_sink == NULL) ? 0
	                                            : tee->tee_sink->id;
}

static void gdam_tee_arg_set_tee_source(GtkObject* object, GdamArg* arg)
{
	GdamTee* tee = GDAM_TEE(object);
	...
}

static void gdam_tee_arg_get_tee_source(GtkObject* object, GdamArg* arg)
{
	GdamTee* tee = GDAM_TEE(object);
	GdamSink* sink;
	if (tee->tee_sink == NULL) {
		GDAM_ARG_INT(arg) = 0;
		return;
	}
	sink = GDAM_SINK(tee->tee_sink);
	if (sink->feed == NULL) {
		GDAM_ARG_INT(arg) = 0;
		return;
	}
	GDAM_ARG_INT(arg) = tee->feed->source_id;
}

static void gdam_tee_class_init(GdamSourceClass* source_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(source_class);
	object_class->destroy = gdam_tee_destroy;

	gdam_arg_register(object_class,
	                  "tee_sink",
			  GDAM_ARG_TYPE_INT,
			  gdam_tee_arg_set_tee_sink,
			  gdam_tee_arg_get_tee_sink,
			  GDAM_ARG_VITAL);

	/* Don't make this vital as it depends on `tee_sink'. */
	gdam_arg_register(object_class,
	                  "tee_source",
			  GDAM_ARG_TYPE_INT,
			  gdam_tee_arg_set_tee_source,
			  gdam_tee_arg_get_tee_source,
			  0);
}

GtkType gdam_tee_get_type()
{
	static GtkType tee_type = 0;
	if (!tee_type) {
		static const GtkTypeInfo tee_info = {
			"GdamTee",
			sizeof(GdamTee),
			sizeof(GdamTeeClass),
			(GtkClassInitFunc) gdam_tee_class_init,
			(GtkObjectInitFunc) gdam_tee_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		tee_type = gtk_type_unique(GDAM_TYPE_SOURCE, &tee_info);
	}
	return tee_type;
}
